////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2004 Xilinx, Inc.  All rights reserved.
//
// Xilinx, Inc.
// XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
// COURTESY TO YOU.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
// ONE POSSIBLE   IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
// STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
// IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
// FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION.
// XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
// THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
// ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
// FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $Id: ppc-hw.c,v 1.4.8.6 2005/11/15 23:41:08 salindac Exp $
//
//////////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------------------------------------------------//
//! @file ppc-hw.c
//! PPC Hardware initialization
//----------------------------------------------------------------------------------------------------//
#include <os_config.h>
#include <config/config_param.h>
#include <config/config_cparam.h>
#include <sys/ktypes.h>
#include <xparameters.h>
#include <sys/process.h>
#include <xexception_l.h>
#include <xtime_l.h>
#include <xpseudo_asm.h>
#include <sys/syscall.h>
#ifdef CONFIG_CACHE
#include <xcache_l.h>
#endif
#ifdef CONFIG_INTC
#include <xintc.h>
#endif
#include <sys/decls.h>

//----------------------------------------------------------------------------------------------------//
// Data
//----------------------------------------------------------------------------------------------------// 
#ifdef CONFIG_INTC
XIntc sys_intc;
#endif
struct t_syscall_data system_call_data;
// this pointer points to data shared between user and system 
struct t_syscall_data *system_call_data_ptr = &system_call_data; 
struct t_syscall_data **system_call_data_ptr_ptr = &system_call_data_ptr;
unsigned int pit_timer_interval = 1; // default interval
extern char did_resched;
extern char timer_need_refresh;

//----------------------------------------------------------------------------------------------------//
// Declarations
//----------------------------------------------------------------------------------------------------//
void hw_init();
void int_system_init();
unsigned int pit_get_value();
void pit_initialize(unsigned int interval);
void pit_set_interval( unsigned int interval);
void pit_reset();
void pit_continue();
void pit_disable();
void setup_initial_context (process_struct *pcb, pid_t parent, unsigned int start_addr, unsigned int stackaddr);
void init_idle_task ();

extern void timer_int_handler(int) ;
extern void system_call_handler(void *baseaddr);

//----------------------------------------------------------------------------------------------------//
// Definitions
//----------------------------------------------------------------------------------------------------//

//----------------------------------------------------------------------------------------------------//
//  @func - hw_init
//! @desc
//!   PPC405 hardware specific initialization
//!   - Enable caches if configured
//!   - Initialize PPC exception handling mechanims
//!   - If an interrupt controller is present register interrupt controller handler 
//!     as the external interrupt handler.
//!   - Register PPC timer interrupt handler as the handler for the PPC PIT interrupt
//! @param
//!   - none
//! @return
//!   - nothing
//! @note
//!   - none
//----------------------------------------------------------------------------------------------------//
void hw_init()
{
    XExc_Init ();                                                                                       // Initialize exception handling 
    XExc_mDisableExceptions (XEXC_NON_CRITICAL);
#ifdef CONFIG_INTC
    int_system_init();
#endif

    XExc_RegisterHandler (XEXC_ID_PIT_INT, (XExceptionHandler)timer_int_handler, (void *)0);            // Register PIT interrupt handler
    pit_initialize (SYSTMR_INTERVAL);                                                                   // use SYSTMR_INTERVAL as configured in MSS
	
    XTime_PITEnableInterrupt();                                                                         // Enable PPC interval timer interrupt
}

#ifdef CONFIG_INTC
//----------------------------------------------------------------------------------------------------//
//  @func - int_system_init
//! @desc
//!   Initialize PPC405 interrupt handling system in the prescence of an interrupt controller.
//! @param
//!   - none
//! @return
//!   - nothing
//! @note
//!   - none
//----------------------------------------------------------------------------------------------------//
void int_system_init ()
{
    XStatus status;
    status = XIntc_Initialize (&sys_intc, SYSINTC_DEVICE_ID);
    if (status != XST_SUCCESS) {
	DBG_PRINT ("XMK: init_int_system: XIntc_Initialize failed.\r\n");
	return;
    }

    XIntc_SetIntrSvcOption(SYSINTC_BASEADDR, XIN_SVC_ALL_ISRS_OPTION);  // IntC ISR should service all interrupts
    XExc_RegisterHandler (XEXC_ID_NON_CRITICAL_INT,                     // Now tie the INTC's interrupt handler to the PPC Non-Critical Interrupt Handler
                          (XExceptionHandler)XIntc_DeviceInterruptHandler, (void*)SYSINTC_DEVICE_ID);   

    status = XIntc_Start (&sys_intc, XIN_REAL_MODE);                    // Start the interrupt controller

    if (status != XST_SUCCESS) {
	DBG_PRINT ("XMK: init_int_system: XIntc_Start failed.\r\n");
	return;
    }
}
#endif /* CONFIG_INTC */

//----------------------------------------------------------------------------------------------------//
//  @func - xmk_enter_kernel
//! @desc
//!   Lock kernel by turning off ALL system interrupts
//! @return
//!   - Nothing
//! @note
//!   - Locks kernel against ALL interrupts.
//!   - Does not lock against critical interrupts.
//----------------------------------------------------------------------------------------------------//
inline void xmk_enter_kernel ()
{
    XExc_mDisableExceptions(XEXC_NON_CRITICAL);                         // Does not disable critical interrupt though
}

//----------------------------------------------------------------------------------------------------//
//  @func - xmk_leave_kernel
//! @desc
//!   Unlock kernel by turning on system timer interrupts
//! @return
//!   - Nothing
//! @note
//!   - Unlocks kernel with respect to ALL interrupts except critical interrupts
//----------------------------------------------------------------------------------------------------//
inline void xmk_leave_kernel( void ) 
{
    XExc_mEnableExceptions(XEXC_NON_CRITICAL);
}



//----------------------------------------------------------------------------------------------------//
//  @func - pit_initialize
//! @desc
//!   Initialize the PIT timer in the system
//! @param
//!   - interval is the interval to interrupt the system at.
//! @return
//!   - nothing
//! @note
//!  - none
//----------------------------------------------------------------------------------------------------//
void pit_initialize (unsigned int interval)
{
    pit_timer_interval = interval;
    XTime_PITDisableInterrupt ();
    XTime_PITSetInterval (interval);	
    XTime_PITEnableAutoReload ();
}

//----------------------------------------------------------------------------------------------------//
//  @func - pit_reset
//! @desc
//!   Enable and start PIT timer with configured PIT interval loaded to the PIT counter
//! @param
//!   - none
//! @return
//!   - nothing
//! @note
//!   - none
//----------------------------------------------------------------------------------------------------//
void pit_reset ()
{
    // Reset only if a new process is scheduled or if the timer ran out
    if (did_resched || timer_need_refresh) {
        XTime_PITSetInterval (pit_timer_interval) ;	
        XTime_PITEnableAutoReload () ;
        XTime_TSRClearStatusBits (XREG_TSR_PIT_INTERRUPT_STATUS);
        XTime_PITEnableInterrupt();
        did_resched = 0;
        timer_need_refresh = 0;
    }
}

void pit_disable ()
{
    XTime_PITDisableInterrupt ();
}


void setup_initial_context (process_struct *pcb, pid_t parent, unsigned int start_addr, unsigned int stackaddr)
{
    pcb->pcontext.isrflag = 1;
    pcb->pcontext.regs[CTX_INDEX_MSR] = ptable[parent].pcontext.regs[CTX_INDEX_MSR] | XEXC_NON_CRITICAL;
    pcb->pcontext.regs[CTX_INDEX_PC] = start_addr;
    pcb->pcontext.regs[CTX_INDEX_GPR(1)] = stackaddr;
    pcb->pcontext.regs[CTX_INDEX_GPR(2)]  = ptable[parent].pcontext.regs[CTX_INDEX_GPR(2)];
    pcb->pcontext.regs[CTX_INDEX_GPR(13)] = ptable[parent].pcontext.regs[CTX_INDEX_GPR(13)];
}

void init_idle_task ()
{
    idle_task_pid = proc_create (PRIO_LOWEST);                          // Idle task (PID 0).
    ptable[idle_task_pid].state = PROC_RUN;                             // Idle task assumed to be running as soon as the kernel starts
    ptable[idle_task_pid].pcontext.regs[CTX_INDEX_MSR]    = mfmsr () | XEXC_NON_CRITICAL;
    ptable[idle_task_pid].pcontext.regs[CTX_INDEX_PC]     = (unsigned int)idle_task;
    ptable[idle_task_pid].pcontext.regs[CTX_INDEX_GPR(1)] = mfgpr (1);
    ptable[idle_task_pid].pcontext.regs[CTX_INDEX_GPR(2)] = mfgpr (2);
    ptable[idle_task_pid].pcontext.regs[CTX_INDEX_GPR(13)]= mfgpr (13);
    SET_CURRENT_PROCESS (idle_task_pid);
}


/* #ifdef CONFIG_PIT_CONTROL */
/* unsigned int pit_get_value () */
/* { */
/*     unsigned int val; */
/*     val = mfspr (XREG_SPR_PIT); */
/*     return val; */
/* } */

/* void pit_set_interval (unsigned int interval) */
/* { */
/*     mtspr(XREG_SPR_PIT, interval); */
/* } */


/* void pit_continue () */
/* { */
/*     // Simulating a continue */
/*     // On the next interrupt due to the pit, the configured interval value  */
/*     // will get loaded onto the PIT */
/*     pit_set_interval (pit_stopped_value); */
/*     XTime_PITEnableInterrupt (); */
/* } */

/* #endif /\* CONFIG_PIT_CONTROL *\/ */

